# -*- coding: utf-8 -*-
#   Copyright 2020, 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Back-end to run quantum program on AQT cloud platform"""

import getpass
import signal
import time

import requests
from requests.compat import urljoin
from requests import Session

# An AQT token can be requested at:
# https://gateway-portal.aqt.eu/

_API_URL = 'https://gateway.aqt.eu/marmot/'


class AQT(Session):
    """Class managing the session to AQT's APIs"""

    def __init__(self):
        super().__init__()
        self.backends = dict()
        self.timeout = 5.0
        self.token = None

    def update_devices_list(self, verbose=False):
        """
        Returns:
            (list): list of available devices

        Up to my knowledge there is no proper API call for online devices,
        so we just assume that the list from AQT portal always up to date
        """
        # TODO: update once the API for getting online devices is available
        self.backends = dict()
        self.backends['aqt_simulator'] = {'nq': 11, 'version': '0.0.1', 'url': 'sim/'}
        self.backends['aqt_simulator_noise'] = {
            'nq': 11,
            'version': '0.0.1',
            'url': 'sim/noise-model-1',
        }
        self.backends['aqt_device'] = {'nq': 4, 'version': '0.0.1', 'url': 'lint/'}
        if verbose:
            print('- List of AQT devices available:')
            print(self.backends)

    def is_online(self, device):
        """
        Check whether a device is currently online

        Args:
            device (str): name of the aqt device to use

        Note:
            Useless at the moment, may change if the API evolves
        """
        return device in self.backends

    def can_run_experiment(self, info, device):
        """
        check if the device is big enough to run the code

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the aqt device to use
        Returns:
            (bool): True if device is big enough, False otherwise
        """
        nb_qubit_max = self.backends[device]['nq']
        nb_qubit_needed = info['nq']
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def authenticate(self, token=None):
        """
        Args:
            token (str): AQT user API token.
        """
        if token is None:
            token = getpass.getpass(prompt='AQT token > ')
        self.headers.update({'Ocp-Apim-Subscription-Key': token, 'SDK': 'ProjectQ'})
        self.token = token

    def run(self, info, device):
        """Run a quantum circuit"""
        argument = {
            'data': info['circuit'],
            'access_token': self.token,
            'repetitions': info['shots'],
            'no_qubits': info['nq'],
        }
        req = super().put(urljoin(_API_URL, self.backends[device]['url']), data=argument)
        req.raise_for_status()
        r_json = req.json()
        if r_json['status'] != 'queued':
            raise Exception('Error in sending the code online')
        execution_id = r_json["id"]
        return execution_id

    def get_result(  # pylint: disable=too-many-arguments
        self, device, execution_id, num_retries=3000, interval=1, verbose=False
    ):
        """
        Get the result of an execution
        """
        if verbose:
            print("Waiting for results. [Job ID: {}]".format(execution_id))

        original_sigint_handler = signal.getsignal(signal.SIGINT)

        def _handle_sigint_during_get_result(*_):  # pragma: no cover
            raise Exception("Interrupted. The ID of your submitted job is {}.".format(execution_id))

        try:
            signal.signal(signal.SIGINT, _handle_sigint_during_get_result)

            for retries in range(num_retries):

                argument = {'id': execution_id, 'access_token': self.token}
                req = super().put(urljoin(_API_URL, self.backends[device]['url']), data=argument)
                req.raise_for_status()
                r_json = req.json()
                if r_json['status'] == 'finished' or 'samples' in r_json:
                    return r_json['samples']
                if r_json['status'] != 'running':
                    raise Exception("Error while running the code: {}.".format(r_json['status']))
                time.sleep(interval)
                if self.is_online(device) and retries % 60 == 0:
                    self.update_devices_list()

                    # TODO: update once the API for getting online devices is
                    #       available
                    if not self.is_online(device):  # pragma: no cover
                        raise DeviceOfflineError(
                            "Device went offline. The ID of your submitted job is {}.".format(execution_id)
                        )

        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

        raise Exception("Timeout. The ID of your submitted job is {}.".format(execution_id))


class DeviceTooSmall(Exception):
    """Exception raised if the device is too small to run the circuit"""


class DeviceOfflineError(Exception):
    """Exception raised if a selected device is currently offline"""


def show_devices(verbose=False):
    """
    Access the list of available devices and their properties (ex: for setup
    configuration)

    Args:
        verbose (bool): If True, additional information is printed

    Returns:
        (list) list of available devices and their properties
    """
    aqt_session = AQT()
    aqt_session.update_devices_list(verbose=verbose)
    return aqt_session.backends


def retrieve(device, token, jobid, num_retries=3000, interval=1, verbose=False):  # pylint: disable=too-many-arguments
    """
    Retrieves a previously run job by its ID.

    Args:
        device (str): Device on which the code was run / is running.
        token (str): AQT user API token.
        jobid (str): Id of the job to retrieve

    Returns:
        (list) samples form the AQT server
    """
    aqt_session = AQT()
    aqt_session.authenticate(token)
    aqt_session.update_devices_list(verbose)
    res = aqt_session.get_result(device, jobid, num_retries=num_retries, interval=interval, verbose=verbose)
    return res


def send(
    info,
    device='aqt_simulator',
    token=None,
    num_retries=100,
    interval=1,
    verbose=False,
):  # pylint: disable=too-many-arguments
    """
    Sends cicruit through the AQT API and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the aqt device. Simulator chosen by default
        token (str): AQT user API token.
        verbose (bool): If True, additional information is printed, such as
            measurement statistics. Otherwise, the backend simply registers
            one measurement result (same behavior as the projectq Simulator).

    Returns:
        (list) samples form the AQT server

    """
    try:
        aqt_session = AQT()

        if verbose:
            print("- Authenticating...")
        if token is not None:
            print('user API token: ' + token)
        aqt_session.authenticate(token)

        # check if the device is online
        aqt_session.update_devices_list(verbose)
        online = aqt_session.is_online(device)
        # useless for the moment
        if not online:  # pragma: no cover
            print("The device is offline (for maintenance?). Use the " "simulator instead or try again later.")
            raise DeviceOfflineError("Device is offline.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = aqt_session.can_run_experiment(info, device)
        if not runnable:
            print(
                "The device is too small ({} qubits available) for the code "
                "requested({} qubits needed). Try to look for another device "
                "with more qubits".format(qmax, qneeded)
            )
            raise DeviceTooSmall("Device is too small.")
        if verbose:
            print("- Running code: {}".format(info))
        execution_id = aqt_session.run(info, device)
        if verbose:
            print("- Waiting for results...")
        res = aqt_session.get_result(
            device,
            execution_id,
            num_retries=num_retries,
            interval=interval,
            verbose=verbose,
        )
        if verbose:
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    except KeyError as err:
        print("- Failed to parse response:")
        print(err)
    return None



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from . import (
    arb1qubit2rzandry,
    barrier,
    carb1qubit2cnotrzandry,
    crz2cxandrz,
    cnot2rxx,
    cnot2cz,
    cnu2toffoliandcu,
    controlstate,
    entangle,
    globalphase,
    h2rx,
    ph2r,
    qubitop2onequbit,
    qft2crandhadamard,
    r2rzandph,
    rx2rz,
    ry2rz,
    rz2rx,
    sqrtswap2cnot,
    stateprep2cnot,
    swap2cnot,
    toffoli2cnotandtgate,
    time_evolution,
    uniformlycontrolledr2cnot,
    phaseestimation,
    amplitudeamplification,
)

all_defined_decomposition_rules = [
    rule
    for module in [
        arb1qubit2rzandry,
        barrier,
        carb1qubit2cnotrzandry,
        crz2cxandrz,
        cnot2rxx,
        cnot2cz,
        cnu2toffoliandcu,
        controlstate,
        entangle,
        globalphase,
        h2rx,
        ph2r,
        qubitop2onequbit,
        qft2crandhadamard,
        r2rzandph,
        rx2rz,
        ry2rz,
        rz2rx,
        sqrtswap2cnot,
        stateprep2cnot,
        swap2cnot,
        toffoli2cnotandtgate,
        time_evolution,
        uniformlycontrolledr2cnot,
        phaseestimation,
        amplitudeamplification,
    ]
    for rule in module.all_defined_decomposition_rules
]



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import pytest

from projectq.backends import Simulator
from projectq.backends._ionq._ionq_mapper import BoundedQubitMapper
from projectq.cengines import MainEngine, DummyEngine
from projectq.meta import LogicalQubitIDTag
from projectq.ops import AllocateQubitGate, Command, DeallocateQubitGate
from projectq.types import WeakQubitRef


def test_cannot_allocate_past_max():
    mapper = BoundedQubitMapper(1)
    engine = MainEngine(
        DummyEngine(),
        engine_list=[mapper],
        verbose=True,
    )
    engine.allocate_qubit()
    with pytest.raises(RuntimeError) as excinfo:
        engine.allocate_qubit()

    assert str(excinfo.value) == "Cannot allocate more than 1 qubits!"

    # Avoid double error reporting
    mapper.current_mapping = {0: 0, 1: 1}


def test_cannot_reallocate_same_qubit():
    engine = MainEngine(
        Simulator(),
        engine_list=[BoundedQubitMapper(1)],
        verbose=True,
    )
    qureg = engine.allocate_qubit()
    qubit = qureg[0]
    qubit_id = qubit.id
    with pytest.raises(RuntimeError) as excinfo:
        allocate_cmd = Command(
            engine=engine,
            gate=AllocateQubitGate(),
            qubits=([WeakQubitRef(engine=engine, idx=qubit_id)],),
            tags=[LogicalQubitIDTag(qubit_id)],
        )
        engine.send([allocate_cmd])

    assert str(excinfo.value) == "Qubit with id 0 has already been allocated!"


def test_cannot_deallocate_unknown_qubit():
    engine = MainEngine(
        Simulator(),
        engine_list=[BoundedQubitMapper(1)],
        verbose=True,
    )
    qureg = engine.allocate_qubit()
    with pytest.raises(RuntimeError) as excinfo:
        deallocate_cmd = Command(
            engine=engine,
            gate=DeallocateQubitGate(),
            qubits=([WeakQubitRef(engine=engine, idx=1)],),
            tags=[LogicalQubitIDTag(1)],
        )
        engine.send([deallocate_cmd])
    assert str(excinfo.value) == "Cannot deallocate a qubit that is not already allocated!"

    # but we can still deallocate an already allocated one
    engine.deallocate_qubit(qureg[0])
    del qureg
    del engine


def test_cannot_deallocate_same_qubit():
    mapper = BoundedQubitMapper(1)
    engine = MainEngine(
        Simulator(),
        engine_list=[mapper],
        verbose=True,
    )
    qureg = engine.allocate_qubit()
    qubit_id = qureg[0].id
    engine.deallocate_qubit(qureg[0])

    with pytest.raises(RuntimeError) as excinfo:
        deallocate_cmd = Command(
            engine=engine,
            gate=DeallocateQubitGate(),
            qubits=([WeakQubitRef(engine=engine, idx=qubit_id)],),
            tags=[LogicalQubitIDTag(qubit_id)],
        )
        engine.send([deallocate_cmd])

    assert str(excinfo.value) == "Cannot deallocate a qubit that is not already allocated!"


def test_flush_deallocates_all_qubits():
    mapper = BoundedQubitMapper(10)
    engine = MainEngine(
        Simulator(),
        engine_list=[mapper],
        verbose=True,
    )
    # needed to prevent GC from removing qubit refs
    qureg = engine.allocate_qureg(10)
    assert len(mapper.current_mapping.keys()) == 10
    assert len(engine.active_qubits) == 10
    engine.flush()
    # Should still be around after flush
    assert len(engine.active_qubits) == 10
    assert len(mapper.current_mapping.keys()) == 10

    # GC will clean things up
    del qureg
    assert len(engine.active_qubits) == 0
    assert len(mapper.current_mapping.keys()) == 0



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setups.restrictedgateset."""

import pytest

import projectq
from projectq.cengines import DummyEngine
from projectq.libs.math import AddConstant, AddConstantModN, MultiplyByConstantModN
from projectq.ops import (
    BasicGate,
    CNOT,
    CRz,
    H,
    Measure,
    QFT,
    QubitOperator,
    Rx,
    Rz,
    Swap,
    TimeEvolution,
    Toffoli,
    X,
)
from projectq.meta import Control

import projectq.setups.restrictedgateset as restrictedgateset


def test_parameter_any():
    engine_list = restrictedgateset.get_engine_list(one_qubit_gates="any", two_qubit_gates="any")
    backend = DummyEngine(save_commands=True)
    eng = projectq.MainEngine(backend, engine_list)
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    gate = BasicGate()
    gate | (qubit1, qubit2)
    gate | qubit1
    eng.flush()
    print(len(backend.received_commands))
    assert backend.received_commands[2].gate == gate
    assert backend.received_commands[3].gate == gate


def test_restriction():
    engine_list = restrictedgateset.get_engine_list(
        one_qubit_gates=(Rz, H),
        two_qubit_gates=(CNOT, AddConstant, Swap),
        other_gates=(Toffoli, AddConstantModN, MultiplyByConstantModN(2, 8)),
    )
    backend = DummyEngine(save_commands=True)
    eng = projectq.MainEngine(backend, engine_list, verbose=True)
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qubit3 = eng.allocate_qubit()
    eng.flush()
    CNOT | (qubit1, qubit2)
    H | qubit1
    with Control(eng, qubit2):
        Rz(0.2) | qubit1
    Measure | qubit1
    AddConstant(1) | (qubit1 + qubit2)
    AddConstantModN(1, 9) | (qubit1 + qubit2 + qubit3)
    Toffoli | (qubit1 + qubit2, qubit3)
    Swap | (qubit1, qubit2)
    MultiplyByConstantModN(2, 8) | qubit1 + qubit2 + qubit3
    TimeEvolution(0.5, QubitOperator("X0 Y1 Z2")) | qubit1 + qubit2 + qubit3
    QFT | qubit1 + qubit2 + qubit3
    Rx(0.1) | (qubit1)
    MultiplyByConstantModN(2, 9) | qubit1 + qubit2 + qubit3
    eng.flush()
    assert backend.received_commands[4].gate == X
    assert len(backend.received_commands[4].control_qubits) == 1
    assert backend.received_commands[5].gate == H
    assert backend.received_commands[6].gate == Rz(0.1)
    assert backend.received_commands[10].gate == Measure
    assert backend.received_commands[11].gate == AddConstant(1)
    assert backend.received_commands[12].gate == AddConstantModN(1, 9)
    assert backend.received_commands[13].gate == X
    assert len(backend.received_commands[13].control_qubits) == 2
    assert backend.received_commands[14].gate == Swap
    assert backend.received_commands[15].gate == MultiplyByConstantModN(2, 8)
    for cmd in backend.received_commands[16:]:
        assert cmd.gate != QFT
        assert not isinstance(cmd.gate, Rx)
        assert not isinstance(cmd.gate, MultiplyByConstantModN)
        assert not isinstance(cmd.gate, TimeEvolution)


def test_wrong_init():
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(two_qubit_gates=(CNOT))
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(one_qubit_gates="Any")
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(other_gates="any")
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(one_qubit_gates=(CRz,))
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(two_qubit_gates=(CRz,))
    with pytest.raises(TypeError):
        restrictedgateset.get_engine_list(other_gates=(CRz,))



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# pylint: skip-file

"""Example of a basic Bernstein-Vazirani circuit using an IonQBackend."""

import getpass
import random

import matplotlib.pyplot as plt

import projectq.setups.ionq
from projectq import MainEngine
from projectq.backends import IonQBackend
from projectq.libs.hist import histogram
from projectq.ops import CX, All, Barrier, H, Measure, Z


def oracle(qureg, input_size, s):
    """Apply the 'oracle'."""

    for bit in range(input_size):
        if s[input_size - 1 - bit] == '1':
            CX | (qureg[bit], qureg[input_size])


def run_bv_circuit(eng, input_size, s_int):
    s = ('{0:0' + str(input_size) + 'b}').format(s_int)
    print("Secret string: ", s)
    print("Number of qubits: ", str(input_size + 1))
    circuit = eng.allocate_qureg(input_size + 1)
    All(H) | circuit
    Z | circuit[input_size]

    Barrier | circuit

    oracle(circuit, input_size, s)

    Barrier | circuit

    qubits = circuit[:input_size]
    All(H) | qubits
    All(Measure) | qubits
    eng.flush()

    # return a random answer from our results
    histogram(eng.backend, qubits)
    plt.show()

    # return a random answer from our results
    probabilities = eng.backend.get_probabilities(qubits)
    random_answer = random.choice(list(probabilities.keys()))
    print("Probability of getting correct string: ", probabilities[s[::-1]])
    return [int(s) for s in random_answer]


if __name__ == '__main__':
    token = None
    device = None
    if token is None:
        token = getpass.getpass(prompt='IonQ apiKey > ')
    if device is None:
        device = input('IonQ device > ')

    # create main compiler engine for the IonQ back-end
    backend = IonQBackend(
        use_hardware=True,
        token=token,
        num_runs=1,
        verbose=False,
        device=device,
    )
    engine_list = projectq.setups.ionq.get_engine_list(
        token=token,
        device=device,
    )
    engine = MainEngine(backend, engine_list)

    # run the circuit and print the result
    print(run_bv_circuit(engine, 3, 3))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Contains the definition of the quantum phase estimation gate"""

from ._basics import BasicGate


class QPE(BasicGate):
    """
    Quantum Phase Estimation gate.

    See setups.decompositions for the complete implementation
    """

    def __init__(self, unitary):
        BasicGate.__init__(self)
        self.unitary = unitary

    def __str__(self):
        return 'QPE({})'.format(str(self.unitary))



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.cnot2cz.py."

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)
from projectq.meta import Control
from projectq.ops import All, CNOT, CZ, Measure, X, Z

from projectq.setups.decompositions import cnot2cz


def test_recognize_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, verbose=True)
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qubit3 = eng.allocate_qubit()
    eng.flush()
    CZ | (qubit1, qubit2)
    with Control(eng, qubit2):
        Z | qubit1
        X | qubit1
    with Control(eng, qubit2 + qubit3):
        Z | qubit1
    eng.flush()  # To make sure gates arrive before deallocate gates
    eng.flush(deallocate_qubits=True)
    # Don't test initial 4 allocate and flush
    for cmd in saving_backend.received_commands[5:7]:
        assert cnot2cz._recognize_cnot(cmd)
    for cmd in saving_backend.received_commands[7:9]:
        assert not cnot2cz._recognize_cnot(cmd)


def _decomp_gates(eng, cmd):
    if len(cmd.control_qubits) == 1 and isinstance(cmd.gate, X.__class__):
        return False
    return True


def test_cnot_decomposition():
    for basis_state_index in range(0, 4):
        basis_state = [0] * 4
        basis_state[basis_state_index] = 1.0
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])
        rule_set = DecompositionRuleSet(modules=[cnot2cz])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(_decomp_gates),
                test_dummy_eng,
            ],
        )
        test_sim = test_eng.backend
        correct_sim = correct_eng.backend
        correct_qb = correct_eng.allocate_qubit()
        correct_ctrl_qb = correct_eng.allocate_qubit()
        correct_eng.flush()
        test_qb = test_eng.allocate_qubit()
        test_ctrl_qb = test_eng.allocate_qubit()
        test_eng.flush()

        correct_sim.set_wavefunction(basis_state, correct_qb + correct_ctrl_qb)
        test_sim.set_wavefunction(basis_state, test_qb + test_ctrl_qb)
        CNOT | (test_ctrl_qb, test_qb)
        CNOT | (correct_ctrl_qb, correct_qb)

        test_eng.flush()
        correct_eng.flush()

        assert len(correct_dummy_eng.received_commands) == 5
        assert len(test_dummy_eng.received_commands) == 7

        for fstate in range(4):
            binary_state = format(fstate, '02b')
            test = test_sim.get_amplitude(binary_state, test_qb + test_ctrl_qb)
            correct = correct_sim.get_amplitude(binary_state, correct_qb + correct_ctrl_qb)
            assert correct == pytest.approx(test, rel=1e-12, abs=1e-12)

        All(Measure) | test_qb + test_ctrl_qb
        All(Measure) | correct_qb + correct_ctrl_qb
        test_eng.flush(deallocate_qubits=True)
        correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setups.decompositions.sqrtswap2cnot."""

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)
from projectq.ops import All, Measure, SqrtSwap, Command
from projectq.types import WeakQubitRef

import projectq.setups.decompositions.sqrtswap2cnot as sqrtswap2cnot


def _decomp_gates(eng, cmd):
    if isinstance(cmd.gate, SqrtSwap.__class__):
        return False
    return True


def test_sqrtswap_invalid():
    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)

    with pytest.raises(ValueError):
        sqrtswap2cnot._decompose_sqrtswap(Command(None, SqrtSwap, ([qb0], [qb1], [qb2])))

    with pytest.raises(ValueError):
        sqrtswap2cnot._decompose_sqrtswap(Command(None, SqrtSwap, ([qb0], [qb1, qb2])))


def test_sqrtswap():
    for basis_state in ([1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]):
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])
        rule_set = DecompositionRuleSet(modules=[sqrtswap2cnot])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(_decomp_gates),
                test_dummy_eng,
            ],
        )
        test_sim = test_eng.backend
        correct_sim = correct_eng.backend
        correct_qureg = correct_eng.allocate_qureg(2)
        correct_eng.flush()
        test_qureg = test_eng.allocate_qureg(2)
        test_eng.flush()

        correct_sim.set_wavefunction(basis_state, correct_qureg)
        test_sim.set_wavefunction(basis_state, test_qureg)

        SqrtSwap | (test_qureg[0], test_qureg[1])
        test_eng.flush()
        SqrtSwap | (correct_qureg[0], correct_qureg[1])
        correct_eng.flush()

        assert len(test_dummy_eng.received_commands) != len(correct_dummy_eng.received_commands)
        for fstate in range(4):
            binary_state = format(fstate, '02b')
            test = test_sim.get_amplitude(binary_state, test_qureg)
            correct = correct_sim.get_amplitude(binary_state, correct_qureg)
            assert correct == pytest.approx(test, rel=1e-10, abs=1e-10)

        All(Measure) | test_qureg
        All(Measure) | correct_qureg
        test_eng.flush(deallocate_qubits=True)
        correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2020, 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.backends._aqt._aqt_http_client.py."""

import pytest
import requests
from requests.compat import urljoin

from projectq.backends._aqt import _aqt_http_client


# Insure that no HTTP request can be made in all tests in this module
@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    monkeypatch.delattr("requests.sessions.Session.request")


_api_url = 'https://gateway.aqt.eu/marmot/'


def test_is_online():
    token = 'access'

    aqt_session = _aqt_http_client.AQT()
    aqt_session.authenticate(token)
    aqt_session.update_devices_list()
    assert aqt_session.is_online('aqt_simulator')
    assert aqt_session.is_online('aqt_simulator_noise')
    assert aqt_session.is_online('aqt_device')
    assert not aqt_session.is_online('aqt_unknown')


def test_show_devices():
    device_list = _aqt_http_client.show_devices(verbose=True)
    # TODO: update once the API for getting online devices is available
    assert len(device_list) == 3


def test_send_too_many_qubits():
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 100,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    token = "access"

    # Code to test:
    with pytest.raises(_aqt_http_client.DeviceTooSmall):
        _aqt_http_client.send(info, device="aqt_simulator", token=token, verbose=True)


def test_send_real_device_online_verbose(monkeypatch):
    json_aqt = {
        'data': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'access_token': 'access',
        'repetitions': 1,
        'no_qubits': 3,
    }
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    token = "access"
    execution_id = '3'
    result_ready = [False]
    result = "my_result"
    request_num = [0]  # To assert correct order of calls

    def mocked_requests_put(*args, **kwargs):
        class MockRequest:
            def __init__(self, body="", url=""):
                self.body = body
                self.url = url

        class MockPutResponse:
            def __init__(self, json_data, text=" "):
                self.json_data = json_data
                self.text = text
                self.request = MockRequest()

            def json(self):
                return self.json_data

            def raise_for_status(self):
                pass

        # Run code
        if args[1] == urljoin(_api_url, "sim/") and kwargs["data"] == json_aqt and request_num[0] == 0:
            request_num[0] += 1
            return MockPutResponse({"id": execution_id, "status": "queued"}, 200)
        elif (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and not result_ready[0]
            and request_num[0] == 1
        ):
            result_ready[0] = True
            request_num[0] += 1
            return MockPutResponse({"status": 'running'}, 200)
        elif (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and result_ready[0]
            and request_num[0] == 2
        ):
            return MockPutResponse({"samples": result, "status": 'finished'}, 200)

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)

    # Code to test:
    res = _aqt_http_client.send(info, device="aqt_simulator", token=None, verbose=True)
    assert res == result


def test_send_that_errors_are_caught(monkeypatch):
    def mocked_requests_put(*args, **kwargs):
        # Test that this error gets caught
        raise requests.exceptions.HTTPError

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)
    # Patch login data
    token = 12345

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    _aqt_http_client.send(info, device="aqt_simulator", token=None, verbose=True)


def test_send_that_errors_are_caught2(monkeypatch):
    def mocked_requests_put(*args, **kwargs):
        # Test that this error gets caught
        raise requests.exceptions.RequestException

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)
    # Patch login data
    token = 12345

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    _aqt_http_client.send(info, device="aqt_simulator", token=None, verbose=True)


def test_send_that_errors_are_caught3(monkeypatch):
    def mocked_requests_put(*args, **kwargs):
        # Test that this error gets caught
        raise KeyError

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)
    # Patch login data
    token = 12345

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    _aqt_http_client.send(info, device="aqt_simulator", token=None, verbose=True)


def test_send_that_errors_are_caught4(monkeypatch):
    json_aqt = {
        'data': '[]',
        'access_token': 'access',
        'repetitions': 1,
        'no_qubits': 3,
    }
    info = {'circuit': '[]', 'nq': 3, 'shots': 1, 'backend': {'name': 'aqt_simulator'}}
    token = "access"
    execution_id = '123e'

    def mocked_requests_put(*args, **kwargs):
        class MockRequest:
            def __init__(self, body="", url=""):
                self.body = body
                self.url = url

        class MockPutResponse:
            def __init__(self, json_data, text=" "):
                self.json_data = json_data
                self.text = text
                self.request = MockRequest()

            def json(self):
                return self.json_data

            def raise_for_status(self):
                pass

        # Run code
        if args[1] == urljoin(_api_url, "sim/") and kwargs["data"] == json_aqt:
            return MockPutResponse({"id": execution_id, "status": "error"}, 200)

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)

    # Code to test:
    _aqt_http_client.time.sleep = lambda x: x
    with pytest.raises(Exception):
        _aqt_http_client.send(
            info,
            device="aqt_simulator",
            token=token,
            num_retries=10,
            verbose=True,
        )


def test_timeout_exception(monkeypatch):
    json_aqt = {
        'data': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'access_token': 'access',
        'repetitions': 1,
        'no_qubits': 3,
    }
    info = {
        'circuit': '[["Y", 0.5, [1]], ["X", 0.5, [1]], ["X", 0.5, [1]], '
        '["Y", 0.5, [1]], ["MS", 0.5, [1, 2]], ["X", 3.5, [1]], '
        '["Y", 3.5, [1]], ["X", 3.5, [2]]]',
        'nq': 3,
        'shots': 1,
        'backend': {'name': 'aqt_simulator'},
    }
    token = "access"
    execution_id = '123e'
    tries = [0]

    def mocked_requests_put(*args, **kwargs):
        class MockRequest:
            def __init__(self, body="", url=""):
                self.body = body
                self.url = url

        class MockPutResponse:
            def __init__(self, json_data, text=" "):
                self.json_data = json_data
                self.text = text
                self.request = MockRequest()

            def json(self):
                return self.json_data

            def raise_for_status(self):
                pass

        # Run code
        if args[1] == urljoin(_api_url, "sim/") and kwargs["data"] == json_aqt:
            return MockPutResponse({"id": execution_id, "status": "queued"}, 200)
        if (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
        ):
            tries[0] += 1
            return MockPutResponse({"status": 'running'}, 200)

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)

    # Code to test:
    _aqt_http_client.time.sleep = lambda x: x
    for tok in (None, token):
        with pytest.raises(Exception) as excinfo:
            _aqt_http_client.send(
                info,
                device="aqt_simulator",
                token=tok,
                num_retries=10,
                verbose=True,
            )
    assert "123e" in str(excinfo.value)  # check that job id is in exception
    assert tries[0] > 0


def test_retrieve(monkeypatch):
    token = "access"
    execution_id = '123e'
    result_ready = [False]
    result = "my_result"
    request_num = [0]  # To assert correct order of calls

    def mocked_requests_put(*args, **kwargs):
        class MockRequest:
            def __init__(self, body="", url=""):
                self.body = body
                self.url = url

        class MockPutResponse:
            def __init__(self, json_data, text=" "):
                self.json_data = json_data
                self.text = text
                self.request = MockRequest()

            def json(self):
                return self.json_data

            def raise_for_status(self):
                pass

        # Run code
        if (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and not result_ready[0]
            and request_num[0] < 1
        ):
            result_ready[0] = True
            request_num[0] += 1
            return MockPutResponse({"status": 'running'}, 200)
        if (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and result_ready[0]
            and request_num[0] == 1
        ):
            return MockPutResponse({"samples": result, "status": 'finished'}, 200)

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)

    def user_password_input(prompt):
        if prompt == "AQT token > ":
            return token

    monkeypatch.setattr("getpass.getpass", user_password_input)

    # Code to test:
    _aqt_http_client.time.sleep = lambda x: x
    res = _aqt_http_client.retrieve(device="aqt_simulator", token=None, verbose=True, jobid="123e")
    assert res == result


def test_retrieve_that_errors_are_caught(monkeypatch):
    token = "access"
    execution_id = '123e'
    result_ready = [False]
    request_num = [0]  # To assert correct order of calls

    def mocked_requests_put(*args, **kwargs):
        class MockRequest:
            def __init__(self, body="", url=""):
                self.body = body
                self.url = url

        class MockPutResponse:
            def __init__(self, json_data, text=" "):
                self.json_data = json_data
                self.text = text
                self.request = MockRequest()

            def json(self):
                return self.json_data

            def raise_for_status(self):
                pass

        # Run code
        if (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and not result_ready[0]
            and request_num[0] < 1
        ):
            result_ready[0] = True
            request_num[0] += 1
            return MockPutResponse({"status": 'running'}, 200)
        if (
            args[1] == urljoin(_api_url, "sim/")
            and kwargs["data"]["access_token"] == token
            and kwargs["data"]["id"] == execution_id
            and result_ready[0]
            and request_num[0] == 1
        ):
            return MockPutResponse({"status": 'error'}, 200)

    monkeypatch.setattr("requests.sessions.Session.put", mocked_requests_put)

    # Code to test:
    _aqt_http_client.time.sleep = lambda x: x
    with pytest.raises(Exception):
        _aqt_http_client.retrieve(device="aqt_simulator", token=token, verbose=True, jobid="123e")



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers a few default replacement rules for Shor's algorithm to work
(see Examples).
"""

from projectq.meta import Control
from projectq.cengines import DecompositionRule

from ._gates import (
    AddConstant,
    AddConstantModN,
    MultiplyByConstantModN,
    AddQuantum,
    SubtractQuantum,
    ComparatorQuantum,
    DivideQuantum,
    MultiplyQuantum,
)

from ._gates import (
    _InverseAddQuantumGate,
    _InverseDivideQuantumGate,
    _InverseMultiplyQuantumGate,
)

from ._constantmath import (
    add_constant,
    add_constant_modN,
    mul_by_constant_modN,
)

from ._quantummath import (
    add_quantum,
    subtract_quantum,
    inverse_add_quantum_carry,
    comparator,
    quantum_conditional_add,
    quantum_division,
    inverse_quantum_division,
    quantum_conditional_add_carry,
    quantum_multiplication,
    inverse_quantum_multiplication,
)


def _replace_addconstant(cmd):
    eng = cmd.engine
    const = cmd.gate.a
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        add_constant(eng, const, quint)


def _replace_addconstmodN(cmd):  # pylint: disable=invalid-name
    eng = cmd.engine
    const = cmd.gate.a
    N = cmd.gate.N
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        add_constant_modN(eng, const, N, quint)


def _replace_multiplybyconstantmodN(cmd):  # pylint: disable=invalid-name
    eng = cmd.engine
    const = cmd.gate.a
    N = cmd.gate.N
    quint = cmd.qubits[0]

    with Control(eng, cmd.control_qubits):
        mul_by_constant_modN(eng, const, N, quint)


def _replace_addquantum(cmd):
    eng = cmd.engine
    if cmd.control_qubits == []:
        quint_a = cmd.qubits[0]
        quint_b = cmd.qubits[1]
        if len(cmd.qubits) == 3:
            carry = cmd.qubits[2]
            add_quantum(eng, quint_a, quint_b, carry)
        else:
            add_quantum(eng, quint_a, quint_b)
    else:
        quint_a = cmd.qubits[0]
        quint_b = cmd.qubits[1]
        if len(cmd.qubits) == 3:
            carry = cmd.qubits[2]
            with Control(eng, cmd.control_qubits):
                quantum_conditional_add_carry(eng, quint_a, quint_b, cmd.control_qubits, carry)
        else:
            with Control(eng, cmd.control_qubits):
                quantum_conditional_add(eng, quint_a, quint_b, cmd.control_qubits)


def _replace_inverse_add_quantum(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]

    if len(cmd.qubits) == 3:
        quint_c = cmd.qubits[2]
        with Control(eng, cmd.control_qubits):
            inverse_add_quantum_carry(eng, quint_a, [quint_b, quint_c])
    else:
        with Control(eng, cmd.control_qubits):
            subtract_quantum(eng, quint_a, quint_b)


def _replace_comparator(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]
    carry = cmd.qubits[2]

    with Control(eng, cmd.control_qubits):
        comparator(eng, quint_a, quint_b, carry)


def _replace_quantumdivision(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]
    quint_c = cmd.qubits[2]

    with Control(eng, cmd.control_qubits):
        quantum_division(eng, quint_a, quint_b, quint_c)


def _replace_inversequantumdivision(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]
    quint_c = cmd.qubits[2]

    with Control(eng, cmd.control_qubits):
        inverse_quantum_division(eng, quint_a, quint_b, quint_c)


def _replace_quantummultiplication(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]
    quint_c = cmd.qubits[2]

    with Control(eng, cmd.control_qubits):
        quantum_multiplication(eng, quint_a, quint_b, quint_c)


def _replace_inversequantummultiplication(cmd):
    eng = cmd.engine
    quint_a = cmd.qubits[0]
    quint_b = cmd.qubits[1]
    quint_c = cmd.qubits[2]

    with Control(eng, cmd.control_qubits):
        inverse_quantum_multiplication(eng, quint_a, quint_b, quint_c)


all_defined_decomposition_rules = [
    DecompositionRule(AddConstant, _replace_addconstant),
    DecompositionRule(AddConstantModN, _replace_addconstmodN),
    DecompositionRule(MultiplyByConstantModN, _replace_multiplybyconstantmodN),
    DecompositionRule(AddQuantum.__class__, _replace_addquantum),
    DecompositionRule(_InverseAddQuantumGate, _replace_inverse_add_quantum),
    DecompositionRule(SubtractQuantum.__class__, _replace_inverse_add_quantum),
    DecompositionRule(ComparatorQuantum.__class__, _replace_comparator),
    DecompositionRule(DivideQuantum.__class__, _replace_quantumdivision),
    DecompositionRule(_InverseDivideQuantumGate, _replace_inversequantumdivision),
    DecompositionRule(MultiplyQuantum.__class__, _replace_quantummultiplication),
    DecompositionRule(_InverseMultiplyQuantumGate, _replace_inversequantummultiplication),
]



# -*- coding: utf-8 -*-
#   Copyright 2020, 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Back-end to run quantum program on AQT's API."""

import math
import random

from projectq.cengines import BasicEngine
from projectq.meta import get_control_count, LogicalQubitIDTag
from projectq.ops import Rx, Ry, Rxx, Measure, Allocate, Barrier, Deallocate, FlushGate
from projectq.types import WeakQubitRef

from ._aqt_http_client import send, retrieve


# _rearrange_result & _format_counts imported and modified from qiskit
def _rearrange_result(input_result, length):
    bin_input = list(bin(input_result)[2:].rjust(length, '0'))
    return ''.join(bin_input)[::-1]


def _format_counts(samples, length):
    counts = {}
    for result in samples:
        h_result = _rearrange_result(result, length)
        if h_result not in counts:
            counts[h_result] = 1
        else:
            counts[h_result] += 1
    return dict(sorted(counts.items(), key=lambda item: item[0]))


class AQTBackend(BasicEngine):  # pylint: disable=too-many-instance-attributes
    """
    The AQT Backend class, which stores the circuit, transforms it to the
    appropriate data format, and sends the circuit through the AQT API.
    """

    def __init__(
        self,
        use_hardware=False,
        num_runs=100,
        verbose=False,
        token='',
        device='simulator',
        num_retries=3000,
        interval=1,
        retrieve_execution=None,
    ):  # pylint: disable=too-many-arguments
        """
        Initialize the Backend object.

        Args:
            use_hardware (bool): If True, the code is run on the AQT quantum
                chip (instead of using the AQT simulator)
            num_runs (int): Number of runs to collect statistics.
                (default is 100, max is usually around 200)
            verbose (bool): If True, statistics are printed, in addition to
                the measurement result being registered (at the end of the
                circuit).
            token (str): AQT user API token.
            device (str): name of the AQT device to use. simulator By default
            num_retries (int): Number of times to retry to obtain
                results from the AQT API. (default is 3000)
            interval (float, int): Number of seconds between successive
                attempts to obtain results from the AQT API.
                (default is 1)
            retrieve_execution (int): Job ID to retrieve instead of re-
                running the circuit (e.g., if previous run timed out).
        """
        BasicEngine.__init__(self)
        self._reset()
        if use_hardware:
            self.device = device
        else:
            self.device = 'simulator'
        self._clear = True
        self._num_runs = num_runs
        self._verbose = verbose
        self._token = token
        self._num_retries = num_retries
        self._interval = interval
        self._probabilities = dict()
        self._circuit = []
        self._mapper = []
        self._measured_ids = []
        self._allocated_qubits = set()
        self._retrieve_execution = retrieve_execution

    def is_available(self, cmd):
        """
        Return true if the command can be executed.

        The AQT ion trap can only do Rx,Ry and Rxx.

        Args:
            cmd (Command): Command for which to check availability
        """
        if get_control_count(cmd) == 0:
            if isinstance(cmd.gate, (Rx, Ry, Rxx)):
                return True
        if cmd.gate in (Measure, Allocate, Deallocate, Barrier):
            return True
        return False

    def _reset(self):
        """Reset all temporary variables (after flush gate)."""
        self._clear = True
        self._measured_ids = []

    def _store(self, cmd):
        """
        Temporarily store the command cmd.

        Translates the command and stores it in a local variable (self._cmds).

        Args:
            cmd: Command to store
        """
        if self._clear:
            self._probabilities = dict()
            self._clear = False
            self._circuit = []
            self._allocated_qubits = set()

        gate = cmd.gate
        if gate == Allocate:
            self._allocated_qubits.add(cmd.qubits[0][0].id)
            return
        if gate == Deallocate:
            return
        if gate == Measure:
            qb_id = cmd.qubits[0][0].id
            logical_id = None
            for tag in cmd.tags:
                if isinstance(tag, LogicalQubitIDTag):
                    logical_id = tag.logical_qubit_id
                    break
            if logical_id is None:
                logical_id = qb_id
                self._mapper.append(qb_id)
            self._measured_ids += [logical_id]
            return
        if isinstance(gate, (Rx, Ry, Rxx)):
            qubits = []
            qubits.append(cmd.qubits[0][0].id)
            if len(cmd.qubits) == 2:
                qubits.append(cmd.qubits[1][0].id)
            angle = gate.angle / math.pi
            instruction = []
            u_name = {'Rx': "X", 'Ry': "Y", 'Rxx': "MS"}
            instruction.append(u_name[str(gate)[0 : int(len(cmd.qubits) + 1)]])  # noqa: E203
            instruction.append(round(angle, 2))
            instruction.append(qubits)
            self._circuit.append(instruction)
            return
        if gate == Barrier:
            return
        raise Exception('Invalid command: ' + str(cmd))

    def _logical_to_physical(self, qb_id):
        """
        Return the physical location of the qubit with the given logical id.
        If no mapper is present then simply returns the qubit ID.

        Args:
            qb_id (int): ID of the logical qubit whose position should be
                returned.
        """
        try:
            mapping = self.main_engine.mapper.current_mapping
            if qb_id not in mapping:
                raise RuntimeError(
                    "Unknown qubit id {}. Please make sure "
                    "eng.flush() was called and that the qubit "
                    "was eliminated during optimization.".format(qb_id)
                )
            return mapping[qb_id]
        except AttributeError as err:
            if qb_id not in self._mapper:
                raise RuntimeError(
                    "Unknown qubit id {}. Please make sure "
                    "eng.flush() was called and that the qubit "
                    "was eliminated during optimization.".format(qb_id)
                ) from err
            return qb_id

    def get_probabilities(self, qureg):
        """
        Return the list of basis states with corresponding probabilities.
        If input qureg is a subset of the register used for the experiment,
        then returns the projected probabilities over the other states.
        The measured bits are ordered according to the supplied quantum
        register, i.e., the left-most bit in the state-string corresponds to
        the first qubit in the supplied quantum register.
        Warning:
            Only call this function after the circuit has been executed!
        Args:
            qureg (list<Qubit>): Quantum register determining the order of the
                qubits.
        Returns:
            probability_dict (dict): Dictionary mapping n-bit strings to
            probabilities.
        Raises:
            RuntimeError: If no data is available (i.e., if the circuit has
                not been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).
        """
        if len(self._probabilities) == 0:
            raise RuntimeError("Please, run the circuit first!")

        probability_dict = dict()
        for state in self._probabilities:
            mapped_state = ['0'] * len(qureg)
            for i, qubit in enumerate(qureg):
                mapped_state[i] = state[self._logical_to_physical(qubit.id)]
            probability = self._probabilities[state]
            mapped_state = "".join(mapped_state)

            probability_dict[mapped_state] = probability_dict.get(mapped_state, 0) + probability
        return probability_dict

    def _run(self):
        """
        Run the circuit.

        Send the circuit via the AQT API using the provided user
        token / ask for the user token.
        """
        # finally: measurements
        # NOTE AQT DOESN'T SEEM TO HAVE MEASUREMENT INSTRUCTIONS (no
        # intermediate measurements are allowed, so implicit at the end)
        # return if no operations.
        if self._circuit == []:
            return

        n_qubit = max(self._allocated_qubits) + 1
        info = {}
        # Hack: AQT instructions specifically need "GATE" string representation
        # instead of 'GATE'
        info['circuit'] = str(self._circuit).replace("'", '"')
        info['nq'] = n_qubit
        info['shots'] = self._num_runs
        info['backend'] = {'name': self.device}
        if self._num_runs > 200:
            raise Exception("Number of shots limited to 200")
        try:
            if self._retrieve_execution is None:
                res = send(
                    info,
                    device=self.device,
                    token=self._token,
                    num_retries=self._num_retries,
                    interval=self._interval,
                    verbose=self._verbose,
                )
            else:
                res = retrieve(
                    device=self.device,
                    token=self._token,
                    jobid=self._retrieve_execution,
                    num_retries=self._num_retries,
                    interval=self._interval,
                    verbose=self._verbose,
                )
            self._num_runs = len(res)
            counts = _format_counts(res, n_qubit)
            # Determine random outcome
            random_outcome = random.random()
            p_sum = 0.0
            measured = ""
            for state in counts:
                probability = counts[state] * 1.0 / self._num_runs
                p_sum += probability
                star = ""
                if p_sum >= random_outcome and measured == "":
                    measured = state
                    star = "*"
                self._probabilities[state] = probability
                if self._verbose and probability > 0:
                    print(str(state) + " with p = " + str(probability) + star)

            # register measurement result from AQT
            for qubit_id in self._measured_ids:
                location = self._logical_to_physical(qubit_id)
                result = int(measured[location])
                self.main_engine.set_measurement_result(WeakQubitRef(self, qubit_id), result)
            self._reset()
        except TypeError as err:
            raise Exception("Failed to run the circuit. Aborting.") from err

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until completion. Upon flush, send the data to the
        AQT API.

        Args:
            command_list: List of commands to execute
        """
        for cmd in command_list:
            if not isinstance(cmd.gate, FlushGate):
                self._store(cmd)
            else:
                self._run()
                self._reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# projectq documentation build configuration file, created by
# sphinx-quickstart on Tue Nov 29 11:51:46 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# pylint: skip-file

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

from importlib.metadata import version

import projectq

# Also import all the modules that are not automatically imported
import projectq.libs.math
import projectq.libs.revkit
import projectq.setups.default
import projectq.setups.grid
import projectq.setups.ibm
import projectq.setups.linear
import projectq.setups.restrictedgateset
import projectq.setups.decompositions

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
import sphinx_rtd_theme

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.autosummary',
    'sphinx.ext.linkcode',
]

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'ProjectQ'
copyright = '2017, ProjectQ'
author = 'ProjectQ'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

release = version('projectq')
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'README.rst']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
# html_title = 'projectq v1'

# A shorter title for the navigation bar.  Default is the same as html_title.
#
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#
# html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x
# 32 pixels large.
#
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#
# html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#
# html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#
# html_additional_pages = {}

# If false, no module index is generated.
#
# html_domain_indices = True

# If false, no index is generated.
#
# html_use_index = True

# If true, the index is split into individual pages for each letter.
#
# html_split_index = False

# If true, links to the reST sources are added to the pages.
#
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr', 'zh'
#
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
#
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#
# html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'projectqdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'projectq.tex', 'projectq Documentation', 'a', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#
# latex_use_parts = False

# If true, show page references after internal links.
#
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
#
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
#
# latex_appendices = []

# It false, will not define \strong, \code,     itleref, \crossref ... but only
# \sphinxstrong, ..., \sphinxtitleref, ... To help avoid clash with user added
# packages.
#
# latex_keep_old_macro_names = True

# If false, no module index is generated.
#
# latex_domain_indices = True

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'projectq', 'projectq Documentation', [author], 1)]

# If true, show URL addresses after external links.
#
# man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'projectq',
        'projectq Documentation',
        author,
        'projectq',
        'One line description of project.',
        'Miscellaneous',
    ),
]

# Documents to append as an appendix to all manuals.
#
# texinfo_appendices = []

# If false, no module index is generated.
#
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#
# texinfo_no_detailmenu = False

# -- Options for sphinx.ext.linkcode --------------------------------------
import inspect


def linkcode_resolve(domain, info):
    # Copyright 2018 ProjectQ (www.projectq.ch), all rights reserved.
    on_rtd = os.environ.get('READTHEDOCS') == 'True'
    github_url = "https://github.com/ProjectQ-Framework/ProjectQ/tree/"
    if on_rtd:
        rtd_tag = os.environ.get('READTHEDOCS_VERSION')
        if rtd_tag == 'latest':
            github_tag = 'develop'
        elif rtd_tag == 'stable':
            github_tag = 'master'
        else:
            # RTD changes "/" in branch name to "-"
            # As we use branches like fix/cool-feature, this is a
            # problem -> as a fix we require that all branch names
            # which contain a '-' must first contain one '/':
            if list(rtd_tag).count('-'):
                github_tag = list(rtd_tag)
                github_tag[github_tag.index('-')] = '/'
                github_tag = ''.join(github_tag)
            else:
                github_tag = rtd_tag
    else:
        github_tag = 'v' + version
    if domain != 'py':
        return None
    else:
        try:
            if 'module' in info and 'fullname' in info and info['module'] and info['fullname']:
                obj = eval(info['module'] + '.' + info['fullname'])
            else:
                return None
        except AttributeError:
            # Object might be a non-static attribute of a class, e.g.,
            # self.num_qubits, which would only exist after init was called.
            # For the moment we don't need a link for that as there is a link
            # for the class already
            return None
        try:
            filepath = inspect.getsourcefile(obj)
            line_number = inspect.getsourcelines(obj)[1]
        except:
            # obj might be a property or a static class variable, e.g.,
            # loop_tag_id in which case obj is an int and inspect will fail
            try:
                # load obj one hierarchy higher (either class or module)
                new_higher_name = info['fullname'].split('.')
                if len(new_higher_name) <= 1:
                    obj = eval(info['module'])
                else:
                    obj = eval(info['module'] + '.' + '.'.join(new_higher_name[:-1]))
                filepath = inspect.getsourcefile(obj)
                line_number = inspect.getsourcelines(obj)[1]
            except:
                return None
        # Only require relative path projectq/relative_path
        projectq_path = inspect.getsourcefile(projectq)[:-11]
        relative_path = os.path.relpath(filepath, projectq_path)
        url = github_url + github_tag + "/projectq/" + relative_path + "#L" + str(line_number)
        return url


# ------------------------------------------------------------------------------

import importlib

sys.path.append(os.path.abspath('.'))
desc = importlib.import_module('package_description')

PackageDescription = desc.PackageDescription

# ------------------------------------------------------------------------------
# Define the description of ProjectQ packages and their submodules below.
#
# In order for the automatic package recognition to work properly, it is
# important that PackageDescription of sub-packages appear earlier in the list
# than their parent package (see for example libs.math and libs.revkit
# compared to libs).
#
# It is also possible to customize the presentation of submodules (see for
# example the setups and setups.decompositions) or even to have private
# sub-modules listed in the documentation page of a parent packages (see for
# example the cengines package)

descriptions = [
    PackageDescription('backends'),
    PackageDescription(
        'cengines',
        desc='''
The ProjectQ compiler engines package.
''',
    ),
    PackageDescription(
        'libs.math',
        desc='''
A tiny math library which will be extended thoughout the next weeks. Right now, it only contains the math functions necessary to run Beauregard's implementation of Shor's algorithm.
''',
    ),
    PackageDescription(
        'libs.revkit',
        desc='''
This library integrates `RevKit <https://msoeken.github.io/revkit.html>`_ into
ProjectQ to allow some automatic synthesis routines for reversible logic.  The
library adds the following operations that can be used to construct quantum
circuits:

- :class:`~projectq.libs.revkit.ControlFunctionOracle`: Synthesizes a reversible circuit from Boolean control function
- :class:`~projectq.libs.revkit.PermutationOracle`: Synthesizes a reversible circuit for a permutation
- :class:`~projectq.libs.revkit.PhaseOracle`: Synthesizes phase circuit from an arbitrary Boolean function

RevKit can be installed from PyPi with `pip install revkit`.

.. note::

    The RevKit Python module must be installed in order to use this ProjectQ library.

    There exist precompiled binaries in PyPi, as well as a source distribution.
    Note that a C++ compiler with C++17 support is required to build the RevKit
    python module from source.  Examples for compatible compilers are Clang
    6.0, GCC 7.3, and GCC 8.1.

The integration of RevKit into ProjectQ and other quantum programming languages is described in the paper

    * Mathias Soeken, Thomas Haener, and Martin Roetteler "Programming Quantum Computers Using Design Automation," in: Design Automation and Test in Europe (2018) [`arXiv:1803.01022 <https://arxiv.org/abs/1803.01022>`_]
''',
        module_special_members='__init__,__or__',
    ),
    PackageDescription(
        'libs',
        desc='''
The library collection of ProjectQ which, for now, consists of a tiny math library and an interface library to RevKit. Soon, more libraries will be added.
''',
    ),
    PackageDescription(
        'meta',
        desc='''
Contains meta statements which allow more optimal code while making it easier for users to write their code.
Examples are `with Compute`, followed by an automatic uncompute or `with Control`, which allows the user to condition an entire code block upon the state of a qubit.
''',
    ),
    PackageDescription(
        'ops',
        desc='''
The operations collection consists of various default gates and is a work-in-progress, as users start to work with ProjectQ.
''',
        module_special_members='__init__,__or__',
    ),
    PackageDescription(
        'setups.decompositions',
        desc='''
The decomposition package is a collection of gate decomposition / replacement rules which can be used by, e.g., the AutoReplacer engine.
''',
    ),
    PackageDescription(
        'setups',
        desc='''
The setups package contains a collection of setups which can be loaded by the `MainEngine`. Each setup contains a `get_engine_list` function which returns a list of compiler engines:

Example:
    .. code-block:: python

        import projectq.setups.ibm as ibm_setup
        from projectq import MainEngine
        eng = MainEngine(engine_list=ibm_setup.get_engine_list())
        # eng uses the default Simulator backend

The subpackage decompositions contains all the individual decomposition rules
which can be given to, e.g., an `AutoReplacer`.
''',
        submodules_desc='''
Each of the submodules contains a setup which can be used to specify the
`engine_list` used by the `MainEngine` :''',
        submodule_special_members='__init__',
    ),
    PackageDescription(
        'types',
        '''
The types package contains quantum types such as Qubit, Qureg, and WeakQubitRef. With further development of the math library, also quantum integers, quantum fixed point numbers etc. will be added.
''',
    ),
]
# ------------------------------------------------------------------------------
# Automatically generate ReST files for each package of ProjectQ

docgen_path = os.path.join(os.path.dirname(os.path.abspath('__file__')), '_doc_gen')
if not os.path.isdir(docgen_path):
    os.mkdir(docgen_path)

for desc in descriptions:
    fname = os.path.join(docgen_path, 'projectq.{}.rst'.format(desc.name))
    lines = None
    if os.path.exists(fname):
        with open(fname, 'r') as fd:
            lines = [line[:-1] for line in fd.readlines()]

    new_lines = desc.get_ReST()

    if new_lines != lines:
        with open(fname, 'w') as fd:
            fd.write('\n'.join(desc.get_ReST()))



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest
from projectq.cengines import BasicEngine, BasicMapperEngine


@pytest.fixture(params=["mapper", "no_mapper"])
def mapper(request):
    """
    Adds a mapper which changes qubit ids by adding 1
    """
    if request.param == "mapper":

        class TrivialMapper(BasicMapperEngine):
            def __init__(self):
                BasicEngine.__init__(self)
                self.current_mapping = dict()

            def receive(self, command_list):
                for cmd in command_list:
                    for qureg in cmd.all_qubits:
                        for qubit in qureg:
                            if qubit.id == -1:
                                continue
                            elif qubit.id not in self.current_mapping:
                                previous_map = self.current_mapping
                                previous_map[qubit.id] = qubit.id + 1
                                self.current_mapping = previous_map
                    self._send_cmd_with_mapped_ids(cmd)

        return TrivialMapper()
    if request.param == "no_mapper":
        return None



# -*- coding: utf-8 -*-
#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setups.decompositions.uniformlycontrolledr2cnot."""

import pytest

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
)

from projectq.meta import Compute, Control, Uncompute
from projectq.ops import (
    All,
    Measure,
    Ry,
    Rz,
    UniformlyControlledRy,
    UniformlyControlledRz,
    X,
)

import projectq.setups.decompositions.uniformlycontrolledr2cnot as ucr2cnot


def slow_implementation(angles, control_qubits, target_qubit, eng, gate_class):
    """
    Assumption is that control_qubits[0] is lowest order bit
    We apply angles[0] to state |0>
    """
    assert len(angles) == 2 ** len(control_qubits)
    for index in range(2 ** len(control_qubits)):
        with Compute(eng):
            for bit_pos in range(len(control_qubits)):
                if not (index >> bit_pos) & 1:
                    X | control_qubits[bit_pos]
        with Control(eng, control_qubits):
            gate_class(angles[index]) | target_qubit
        Uncompute(eng)


def _decomp_gates(eng, cmd):
    if isinstance(cmd.gate, UniformlyControlledRy) or isinstance(cmd.gate, UniformlyControlledRz):
        return False
    return True


def test_no_control_qubits():
    rule_set = DecompositionRuleSet(modules=[ucr2cnot])
    eng = MainEngine(
        backend=DummyEngine(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(_decomp_gates)],
    )
    qb = eng.allocate_qubit()
    with pytest.raises(TypeError):
        UniformlyControlledRy([0.1]) | qb


def test_wrong_number_of_angles():
    rule_set = DecompositionRuleSet(modules=[ucr2cnot])
    eng = MainEngine(
        backend=DummyEngine(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(_decomp_gates)],
    )
    qb = eng.allocate_qubit()
    with pytest.raises(ValueError):
        UniformlyControlledRy([0.1, 0.2]) | ([], qb)


@pytest.mark.parametrize("gate_classes", [(Ry, UniformlyControlledRy), (Rz, UniformlyControlledRz)])
@pytest.mark.parametrize("n", [0, 1, 2, 3, 4])
def test_uniformly_controlled_ry(n, gate_classes):
    random_angles = [
        0.5,
        0.8,
        1.2,
        2.5,
        4.4,
        2.32,
        6.6,
        15.12,
        1,
        2,
        9.54,
        2.1,
        3.1415,
        1.1,
        0.01,
        0.99,
    ]
    angles = random_angles[: 2 ** n]
    for basis_state_index in range(0, 2 ** (n + 1)):
        basis_state = [0] * 2 ** (n + 1)
        basis_state[basis_state_index] = 1.0
        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])
        rule_set = DecompositionRuleSet(modules=[ucr2cnot])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(_decomp_gates),
                test_dummy_eng,
            ],
        )
        test_sim = test_eng.backend
        correct_sim = correct_eng.backend
        correct_qb = correct_eng.allocate_qubit()
        correct_ctrl_qureg = correct_eng.allocate_qureg(n)
        correct_eng.flush()
        test_qb = test_eng.allocate_qubit()
        test_ctrl_qureg = test_eng.allocate_qureg(n)
        test_eng.flush()

        correct_sim.set_wavefunction(basis_state, correct_qb + correct_ctrl_qureg)
        test_sim.set_wavefunction(basis_state, test_qb + test_ctrl_qureg)

        gate_classes[1](angles) | (test_ctrl_qureg, test_qb)
        slow_implementation(
            angles=angles,
            control_qubits=correct_ctrl_qureg,
            target_qubit=correct_qb,
            eng=correct_eng,
            gate_class=gate_classes[0],
        )
        test_eng.flush()
        correct_eng.flush()

        for fstate in range(2 ** (n + 1)):
            binary_state = format(fstate, '0' + str(n + 1) + 'b')
            test = test_sim.get_amplitude(binary_state, test_qb + test_ctrl_qureg)
            correct = correct_sim.get_amplitude(binary_state, correct_qb + correct_ctrl_qureg)
            assert correct == pytest.approx(test, rel=1e-10, abs=1e-10)

        All(Measure) | test_qb + test_ctrl_qureg
        All(Measure) | correct_qb + correct_ctrl_qureg
        test_eng.flush(deallocate_qubits=True)
        correct_eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.time_evolution."
import copy

import numpy
import pytest
import scipy
from scipy import sparse as sps
import scipy.sparse.linalg

from projectq import MainEngine
from projectq.backends import Simulator
from projectq.cengines import (
    DummyEngine,
    AutoReplacer,
    InstructionFilter,
    DecompositionRuleSet,
)
from projectq.meta import Control
from projectq.ops import QubitOperator, TimeEvolution, ClassicalInstructionGate, Ph, Rx, Ry, All, Measure, Command
from projectq.types import WeakQubitRef

from . import time_evolution as te


def test_recognize_commuting_terms():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    wavefunction = eng.allocate_qureg(5)
    op1 = QubitOperator("X1 Y2", 0.5)
    op2 = QubitOperator("Y2 X4", -0.5)
    op3 = QubitOperator((), 0.5)
    op4 = QubitOperator("X1 Y2", 0.5) + QubitOperator("X2", 1e-10)
    op5 = QubitOperator("X1 Y2", 0.5) + QubitOperator("X2", 1e-8)
    op6 = QubitOperator("X2", 1.0)
    TimeEvolution(1.0, op1 + op2 + op3 + op4) | wavefunction
    TimeEvolution(1.0, op1 + op5) | wavefunction
    TimeEvolution(1.0, op1 + op6) | wavefunction
    TimeEvolution(1.0, op1) | wavefunction

    cmd1 = saving_backend.received_commands[5]
    cmd2 = saving_backend.received_commands[6]
    cmd3 = saving_backend.received_commands[7]
    cmd4 = saving_backend.received_commands[8]

    assert te.rule_commuting_terms.gate_recognizer(cmd1)
    assert not te.rule_commuting_terms.gate_recognizer(cmd2)
    assert not te.rule_commuting_terms.gate_recognizer(cmd3)
    assert not te.rule_commuting_terms.gate_recognizer(cmd4)


def test_decompose_commuting_terms():
    saving_backend = DummyEngine(save_commands=True)

    def my_filter(self, cmd):
        if len(cmd.qubits[0]) <= 2 or isinstance(cmd.gate, ClassicalInstructionGate):
            return True
        return False

    rules = DecompositionRuleSet([te.rule_commuting_terms])
    replacer = AutoReplacer(rules)
    filter_eng = InstructionFilter(my_filter)
    eng = MainEngine(backend=saving_backend, engine_list=[replacer, filter_eng])
    qureg = eng.allocate_qureg(5)
    with Control(eng, qureg[3]):
        op1 = QubitOperator("X1 Y2", 0.7)
        op2 = QubitOperator("Y2 X4", -0.8)
        op3 = QubitOperator((), 0.6)
        TimeEvolution(1.5, op1 + op2 + op3) | qureg

    cmd1 = saving_backend.received_commands[5]
    cmd2 = saving_backend.received_commands[6]
    cmd3 = saving_backend.received_commands[7]

    found = [False, False, False]
    scaled_op1 = QubitOperator("X0 Y1", 0.7)
    scaled_op2 = QubitOperator("Y0 X1", -0.8)
    for cmd in [cmd1, cmd2, cmd3]:
        if (
            cmd.gate == Ph(-1.5 * 0.6)
            and cmd.qubits[0][0].id == qureg[1].id
            and cmd.control_qubits[0].id == qureg[3].id  # 1st qubit of [1,2,4]
        ):
            found[0] = True
        elif (
            isinstance(cmd.gate, TimeEvolution)
            and cmd.gate.hamiltonian.isclose(scaled_op1)
            and cmd.gate.time == pytest.approx(1.5)
            and cmd.qubits[0][0].id == qureg[1].id
            and cmd.qubits[0][1].id == qureg[2].id
            and cmd.control_qubits[0].id == qureg[3].id
        ):
            found[1] = True
        elif (
            isinstance(cmd.gate, TimeEvolution)
            and cmd.gate.hamiltonian.isclose(scaled_op2)
            and cmd.gate.time == pytest.approx(1.5)
            and cmd.qubits[0][0].id == qureg[2].id
            and cmd.qubits[0][1].id == qureg[4].id
            and cmd.control_qubits[0].id == qureg[3].id
        ):
            found[2] = True
    assert all(found)


def test_recognize_individual_terms():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend, engine_list=[])
    wavefunction = eng.allocate_qureg(5)
    op1 = QubitOperator("X1 Y2", 0.5)
    op2 = QubitOperator("Y2 X4", -0.5)
    op3 = QubitOperator("X2", 1.0)
    TimeEvolution(1.0, op1 + op2) | wavefunction
    TimeEvolution(1.0, op2) | wavefunction
    TimeEvolution(1.0, op3) | wavefunction

    cmd1 = saving_backend.received_commands[5]
    cmd2 = saving_backend.received_commands[6]
    cmd3 = saving_backend.received_commands[7]

    assert not te.rule_individual_terms.gate_recognizer(cmd1)
    assert te.rule_individual_terms.gate_recognizer(cmd2)
    assert te.rule_individual_terms.gate_recognizer(cmd3)


def test_decompose_individual_terms_invalid():
    eng = MainEngine(backend=DummyEngine(), engine_list=[])
    qb0 = WeakQubitRef(eng, idx=0)
    qb1 = WeakQubitRef(eng, idx=1)
    op1 = QubitOperator("X0 Y1", 0.5)
    op2 = op1 + QubitOperator("Y2 X4", -0.5)
    op3 = QubitOperator(tuple(), 0.5)
    op4 = QubitOperator("X0 Y0", 0.5)

    with pytest.raises(ValueError):
        te._decompose_time_evolution_individual_terms(Command(eng, TimeEvolution(1, op1), ([qb0], [qb1])))

    with pytest.raises(ValueError):
        te._decompose_time_evolution_individual_terms(Command(eng, TimeEvolution(1, op2), ([qb0],)))

    with pytest.raises(ValueError):
        te._decompose_time_evolution_individual_terms(Command(eng, TimeEvolution(1, op3), ([qb0],)))

    with pytest.raises(ValueError):
        te._decompose_time_evolution_individual_terms(Command(eng, TimeEvolution(1, op4), ([qb0, qb1],)))


def test_decompose_individual_terms():
    saving_eng = DummyEngine(save_commands=True)

    def my_filter(self, cmd):
        if isinstance(cmd.gate, TimeEvolution):
            return False
        return True

    rules = DecompositionRuleSet([te.rule_individual_terms])
    replacer = AutoReplacer(rules)
    filter_eng = InstructionFilter(my_filter)
    eng = MainEngine(backend=Simulator(), engine_list=[replacer, filter_eng, saving_eng])
    qureg = eng.allocate_qureg(5)
    # initialize in random wavefunction by applying some gates:
    Rx(0.1) | qureg[0]
    Ry(0.2) | qureg[1]
    Rx(0.45) | qureg[2]
    Rx(0.6) | qureg[3]
    Ry(0.77) | qureg[4]
    eng.flush()
    # Use cheat to get initial start wavefunction:
    qubit_to_bit_map, init_wavefunction = copy.deepcopy(eng.backend.cheat())
    # Apply one qubit gates:
    op1 = QubitOperator((), 0.6)
    op2 = QubitOperator("X2", 0.21)
    op3 = QubitOperator("Y1", 0.33)
    op4 = QubitOperator("Z3", 0.42)
    op5 = QubitOperator("X0 Y1 Z2 Z4", -0.5)
    TimeEvolution(1.1, op1) | qureg
    eng.flush()
    qbit_to_bit_map1, final_wavefunction1 = copy.deepcopy(eng.backend.cheat())
    TimeEvolution(1.2, op2) | qureg
    eng.flush()
    qbit_to_bit_map2, final_wavefunction2 = copy.deepcopy(eng.backend.cheat())
    TimeEvolution(1.3, op3) | qureg
    eng.flush()
    qbit_to_bit_map3, final_wavefunction3 = copy.deepcopy(eng.backend.cheat())
    TimeEvolution(1.4, op4) | qureg
    eng.flush()
    qbit_to_bit_map4, final_wavefunction4 = copy.deepcopy(eng.backend.cheat())
    TimeEvolution(1.5, op5) | qureg
    eng.flush()
    qbit_to_bit_map5, final_wavefunction5 = copy.deepcopy(eng.backend.cheat())
    All(Measure) | qureg

    # Check manually:

    def build_matrix(list_single_matrices):
        res = list_single_matrices[0]
        for i in range(1, len(list_single_matrices)):
            res = sps.kron(res, list_single_matrices[i])
        return res.tocsc()

    id_sp = sps.identity(2, format="csc", dtype=complex)
    x_sp = sps.csc_matrix([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
    y_sp = sps.csc_matrix([[0.0, -1.0j], [1.0j, 0.0]], dtype=complex)
    z_sp = sps.csc_matrix([[1.0, 0.0], [0.0, -1.0]], dtype=complex)

    matrix1 = sps.identity(2 ** 5, format="csc", dtype=complex) * 0.6 * 1.1 * -1.0j
    step1 = scipy.sparse.linalg.expm(matrix1).dot(init_wavefunction)
    assert numpy.allclose(step1, final_wavefunction1)

    matrix2_list = []
    for i in range(5):
        if i == qbit_to_bit_map2[qureg[2].id]:
            matrix2_list.append(x_sp)
        else:
            matrix2_list.append(id_sp)
    matrix2_list.reverse()
    matrix2 = build_matrix(matrix2_list) * 0.21 * 1.2 * -1.0j
    step2 = scipy.sparse.linalg.expm(matrix2).dot(step1)
    assert numpy.allclose(step2, final_wavefunction2)

    matrix3_list = []
    for i in range(5):
        if i == qbit_to_bit_map3[qureg[1].id]:
            matrix3_list.append(y_sp)
        else:
            matrix3_list.append(id_sp)
    matrix3_list.reverse()
    matrix3 = build_matrix(matrix3_list) * 0.33 * 1.3 * -1.0j
    step3 = scipy.sparse.linalg.expm(matrix3).dot(final_wavefunction2)
    assert numpy.allclose(step3, final_wavefunction3)

    matrix4_list = []
    for i in range(5):
        if i == qbit_to_bit_map4[qureg[3].id]:
            matrix4_list.append(z_sp)
        else:
            matrix4_list.append(id_sp)
    matrix4_list.reverse()
    matrix4 = build_matrix(matrix4_list) * 0.42 * 1.4 * -1.0j
    step4 = scipy.sparse.linalg.expm(matrix4).dot(final_wavefunction3)
    assert numpy.allclose(step4, final_wavefunction4)

    matrix5_list = []
    for i in range(5):
        if i == qbit_to_bit_map5[qureg[0].id]:
            matrix5_list.append(x_sp)
        elif i == qbit_to_bit_map5[qureg[1].id]:
            matrix5_list.append(y_sp)
        elif i == qbit_to_bit_map5[qureg[2].id]:
            matrix5_list.append(z_sp)
        elif i == qbit_to_bit_map5[qureg[4].id]:
            matrix5_list.append(z_sp)
        else:
            matrix5_list.append(id_sp)
    matrix5_list.reverse()
    matrix5 = build_matrix(matrix5_list) * -0.5 * 1.5 * -1.0j
    step5 = scipy.sparse.linalg.expm(matrix5).dot(final_wavefunction4)
    print(step5)
    print(final_wavefunction5)

    assert numpy.allclose(step5, final_wavefunction5)



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.cengines._swapandcnotflipper.py."""

import pytest

from projectq import MainEngine
from projectq.cengines import DummyEngine
from projectq.ops import All, H, CNOT, X, Swap, Command
from projectq.meta import Control, Compute, Uncompute, ComputeTag, UncomputeTag
from projectq.types import WeakQubitRef

from . import _swapandcnotflipper


def test_swapandcnotflipper_missing_connection():
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(set())
    eng = MainEngine(DummyEngine(save_commands=True), [flipper])
    qubit1, qubit2 = eng.allocate_qureg(2)
    with pytest.raises(RuntimeError):
        Swap | (qubit1, qubit2)


def test_swapandcnotflipper_invalid_swap():
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(set())

    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)
    with pytest.raises(RuntimeError):
        flipper.receive([Command(engine=None, gate=Swap, qubits=([qb0, qb1], [qb2]))])


def test_swapandcnotflipper_is_available():
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(set())
    dummy = DummyEngine()
    dummy.is_available = lambda x: False
    flipper.next_engine = dummy
    eng = MainEngine(DummyEngine(save_commands=True), [])
    qubit1, qubit2 = eng.allocate_qureg(2)
    Swap | (qubit1, qubit2)
    swap_count = 0
    for cmd in eng.backend.received_commands:
        if cmd.gate == Swap:
            swap_count += 1
            assert flipper.is_available(cmd)
    assert swap_count == 1

    eng = MainEngine(DummyEngine(save_commands=True), [])
    qubit1, qubit2, qubit3 = eng.allocate_qureg(3)
    with Control(eng, qubit3):
        Swap | (qubit1, qubit2)
    swap_count = 0
    for cmd in eng.backend.received_commands:
        if cmd.gate == Swap:
            swap_count += 1
            assert not flipper.is_available(cmd)
    assert swap_count == 1


def test_swapandcnotflipper_flips_cnot():
    backend = DummyEngine(save_commands=True)
    connectivity = set([(0, 1)])
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(connectivity)
    eng = MainEngine(backend=backend, engine_list=[flipper])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    CNOT | (qb0, qb1)
    CNOT | (qb1, qb0)
    hgates = 0
    for cmd in backend.received_commands:
        if cmd.gate == H:
            hgates += 1
        if cmd.gate == X:
            assert cmd.qubits[0][0].id == 1
            assert cmd.control_qubits[0].id == 0
    assert hgates == 4


def test_swapandcnotflipper_invalid_circuit():
    backend = DummyEngine(save_commands=True)
    connectivity = set([(0, 2)])
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(connectivity)
    eng = MainEngine(backend=backend, engine_list=[flipper])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    qb2 = eng.allocate_qubit()
    CNOT | (qb0, qb2)
    CNOT | (qb2, qb0)
    with pytest.raises(RuntimeError):
        CNOT | (qb0, qb1)
    with pytest.raises(RuntimeError):
        Swap | (qb0, qb1)


def test_swapandcnotflipper_optimize_swaps():
    backend = DummyEngine(save_commands=True)
    connectivity = set([(1, 0)])
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(connectivity)
    eng = MainEngine(backend=backend, engine_list=[flipper])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    Swap | (qb0, qb1)
    hgates = 0
    for cmd in backend.received_commands:
        if cmd.gate == H:
            hgates += 1
        if cmd.gate == X:
            assert cmd.qubits[0][0].id == 0
            assert cmd.control_qubits[0].id == 1
    assert hgates == 4

    backend = DummyEngine(save_commands=True)
    connectivity = set([(0, 1)])
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(connectivity)
    eng = MainEngine(backend=backend, engine_list=[flipper])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    Swap | (qb0, qb1)
    hgates = 0
    for cmd in backend.received_commands:
        if cmd.gate == H:
            hgates += 1
        if cmd.gate == X:
            assert cmd.qubits[0][0].id == 1
            assert cmd.control_qubits[0].id == 0
    assert hgates == 4


def test_swapandcnotflipper_keeps_tags():
    backend = DummyEngine(save_commands=True)
    connectivity = set([(1, 0)])
    flipper = _swapandcnotflipper.SwapAndCNOTFlipper(connectivity)
    eng = MainEngine(backend=backend, engine_list=[flipper])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    with Compute(eng):
        All(H) | (qb0 + qb1)
        CNOT | (qb0, qb1)
        CNOT | (qb1, qb0)
        Swap | (qb0, qb1)
    Uncompute(eng)
    hgates = 0
    for cmd in backend.received_commands:
        if cmd.gate == H:
            for t in cmd.tags:
                if isinstance(t, (ComputeTag, UncomputeTag)):
                    hgates += 1
                    break
    assert hgates == 20



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
The projectq.meta package features meta instructions which help both the user and the compiler in writing/producing
efficient code. It includes, e.g.,

* Loop (with Loop(eng): ...)
* Compute/Uncompute (with Compute(eng): ..., [...], Uncompute(eng))
* Control (with Control(eng, ctrl_qubits): ...)
* Dagger (with Dagger(eng): ...)
"""

from ._dirtyqubit import DirtyQubitTag
from ._loop import LoopTag, Loop
from ._compute import Compute, Uncompute, CustomUncompute, ComputeTag, UncomputeTag
from ._control import Control, get_control_count, has_negative_control, canonical_ctrl_state
from ._dagger import Dagger
from ._util import insert_engine, drop_engine_after
from ._logicalqubit import LogicalQubitIDTag



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Test for projectq.backends._awsbraket._awsbraket.py"""

import pytest

import copy
import math
from projectq import MainEngine


from projectq.types import WeakQubitRef
from projectq.cengines import (
    BasicMapperEngine,
    DummyEngine,
    AutoReplacer,
    DecompositionRuleSet,
)
from projectq.cengines._replacer import NoGateDecompositionError

from projectq.ops import (
    R,
    Swap,
    H,
    Rx,
    Ry,
    Rz,
    S,
    Sdag,
    T,
    Tdag,
    X,
    Y,
    Z,
    CNOT,
    SqrtX,
    MatrixGate,
    Entangle,
    Ph,
    NOT,
    C,
    Measure,
    Allocate,
    Deallocate,
    Barrier,
    All,
    Command,
)

from ._awsbraket_test_fixtures import *  # noqa: F401,F403

# ==============================================================================

_has_boto3 = True
try:
    import botocore
    from projectq.backends._awsbraket import _awsbraket
except ImportError:
    _has_boto3 = False

has_boto3 = pytest.mark.skipif(not _has_boto3, reason="boto3 package is not installed")

# ==============================================================================


@pytest.fixture(params=["mapper", "no_mapper"])
def mapper(request):
    """
    Adds a mapper which changes qubit ids by adding 1
    """
    if request.param == "mapper":

        class TrivialMapper(BasicMapperEngine):
            def __init__(self):
                super().__init__()
                self.current_mapping = dict()

            def receive(self, command_list):
                for cmd in command_list:
                    for qureg in cmd.all_qubits:
                        for qubit in qureg:
                            if qubit.id == -1:
                                continue
                            elif qubit.id not in self.current_mapping:
                                previous_map = self.current_mapping
                                previous_map[qubit.id] = qubit.id
                                self.current_mapping = previous_map
                    self._send_cmd_with_mapped_ids(cmd)

        return TrivialMapper()
    if request.param == "no_mapper":
        return None


# ==============================================================================
'''
Gate availability Tests
'''


@has_boto3
@pytest.mark.parametrize(
    "single_qubit_gate_aspen, is_available_aspen",
    [
        (X, True),
        (Y, True),
        (Z, True),
        (H, True),
        (T, True),
        (Tdag, True),
        (S, True),
        (Sdag, True),
        (Allocate, True),
        (Deallocate, True),
        (SqrtX, False),
        (Measure, True),
        (Rx(0.5), True),
        (Ry(0.5), True),
        (Rz(0.5), True),
        (Ph(0.5), False),
        (R(0.5), True),
        (Barrier, True),
        (Entangle, False),
    ],
)
def test_awsbraket_backend_is_available_aspen(single_qubit_gate_aspen, is_available_aspen):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='Aspen-8')
    cmd = Command(eng, single_qubit_gate_aspen, (qubit1,))
    assert aws_backend.is_available(cmd) == is_available_aspen


@has_boto3
@pytest.mark.parametrize(
    "single_qubit_gate_ionq, is_available_ionq",
    [
        (X, True),
        (Y, True),
        (Z, True),
        (H, True),
        (T, True),
        (Tdag, True),
        (S, True),
        (Sdag, True),
        (Allocate, True),
        (Deallocate, True),
        (SqrtX, True),
        (Measure, True),
        (Rx(0.5), True),
        (Ry(0.5), True),
        (Rz(0.5), True),
        (Ph(0.5), False),
        (R(0.5), False),
        (Barrier, True),
        (Entangle, False),
    ],
)
def test_awsbraket_backend_is_available_ionq(single_qubit_gate_ionq, is_available_ionq):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='IonQ Device')
    cmd = Command(eng, single_qubit_gate_ionq, (qubit1,))
    assert aws_backend.is_available(cmd) == is_available_ionq


@has_boto3
@pytest.mark.parametrize(
    "single_qubit_gate_sv1, is_available_sv1",
    [
        (X, True),
        (Y, True),
        (Z, True),
        (H, True),
        (T, True),
        (Tdag, True),
        (S, True),
        (Sdag, True),
        (Allocate, True),
        (Deallocate, True),
        (SqrtX, True),
        (Measure, True),
        (Rx(0.5), True),
        # use MatrixGate as unitary gate
        (MatrixGate([[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]), False),
        (Ry(0.5), True),
        (Rz(0.5), True),
        (Ph(0.5), False),
        (R(0.5), True),
        (Barrier, True),
        (Entangle, False),
    ],
)
def test_awsbraket_backend_is_available_sv1(single_qubit_gate_sv1, is_available_sv1):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=False)
    cmd = Command(eng, single_qubit_gate_sv1, (qubit1,))
    assert aws_backend.is_available(cmd) == is_available_sv1


@has_boto3
@pytest.mark.parametrize(
    "num_ctrl_qubits_aspen, is_available_aspen",
    [(0, True), (1, True), (2, True), (3, False)],
)
def test_awsbraket_backend_is_available_control_not_aspen(num_ctrl_qubits_aspen, is_available_aspen):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(num_ctrl_qubits_aspen)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='Aspen-8')
    cmd = Command(eng, X, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_aspen


@has_boto3
@pytest.mark.parametrize(
    "num_ctrl_qubits_ionq, is_available_ionq",
    [(0, True), (1, True), (2, False), (3, False)],
)
def test_awsbraket_backend_is_available_control_not_ionq(num_ctrl_qubits_ionq, is_available_ionq):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(num_ctrl_qubits_ionq)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='IonQ Device')
    cmd = Command(eng, X, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_ionq


@has_boto3
@pytest.mark.parametrize(
    "num_ctrl_qubits_sv1, is_available_sv1",
    [(0, True), (1, True), (2, True), (3, False)],
)
def test_awsbraket_backend_is_available_control_not_sv1(num_ctrl_qubits_sv1, is_available_sv1):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(num_ctrl_qubits_sv1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=False)
    cmd = Command(eng, X, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_sv1


@has_boto3
@pytest.mark.parametrize(
    "ctrl_singlequbit_aspen, is_available_aspen",
    [
        (X, True),
        (Y, False),
        (Z, True),
        (R(0.5), True),
        (Rx(0.5), False),
        (Ry(0.5), False),
        (Rz(0.5), False),
        (NOT, True),
    ],
)
def test_awsbraket_backend_is_available_control_singlequbit_aspen(ctrl_singlequbit_aspen, is_available_aspen):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='Aspen-8')
    cmd = Command(eng, ctrl_singlequbit_aspen, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_aspen


@has_boto3
@pytest.mark.parametrize(
    "ctrl_singlequbit_ionq, is_available_ionq",
    [
        (X, True),
        (Y, False),
        (Z, False),
        (R(0.5), False),
        (Rx(0.5), False),
        (Ry(0.5), False),
        (Rz(0.5), False),
    ],
)
def test_awsbraket_backend_is_available_control_singlequbit_ionq(ctrl_singlequbit_ionq, is_available_ionq):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='IonQ Device')
    cmd = Command(eng, ctrl_singlequbit_ionq, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_ionq


@has_boto3
@pytest.mark.parametrize(
    "ctrl_singlequbit_sv1, is_available_sv1",
    [
        (X, True),
        (Y, True),
        (Z, True),
        (R(0.5), True),
        (Rx(0.5), False),
        (Ry(0.5), False),
        (Rz(0.5), False),
    ],
)
def test_awsbraket_backend_is_available_control_singlequbit_sv1(ctrl_singlequbit_sv1, is_available_sv1):
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=False)
    cmd = Command(eng, ctrl_singlequbit_sv1, (qubit1,), controls=qureg)
    assert aws_backend.is_available(cmd) == is_available_sv1


def test_awsbraket_backend_is_available_negative_control():
    backend = _awsbraket.AWSBraketBackend()

    qb0 = WeakQubitRef(engine=None, idx=0)
    qb1 = WeakQubitRef(engine=None, idx=1)
    qb2 = WeakQubitRef(engine=None, idx=2)

    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1]))
    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1], control_state='1'))
    assert not backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1], control_state='0'))

    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2]))
    assert backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2], control_state='11'))
    assert not backend.is_available(Command(None, X, qubits=([qb0],), controls=[qb1, qb2], control_state='01'))


@has_boto3
def test_awsbraket_backend_is_available_swap_aspen():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='Aspen-8')
    cmd = Command(eng, Swap, (qubit1, qubit2))
    assert aws_backend.is_available(cmd)


@has_boto3
def test_awsbraket_backend_is_available_swap_ionq():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='IonQ Device')
    cmd = Command(eng, Swap, (qubit1, qubit2))
    assert aws_backend.is_available(cmd)


@has_boto3
def test_awsbraket_backend_is_available_swap_sv1():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=False)
    cmd = Command(eng, Swap, (qubit1, qubit2))
    assert aws_backend.is_available(cmd)


@has_boto3
def test_awsbraket_backend_is_available_control_swap_aspen():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=True, device='Aspen-8')
    cmd = Command(eng, Swap, (qubit1, qubit2), controls=qureg)
    assert aws_backend.is_available(cmd)


@has_boto3
def test_awsbraket_backend_is_available_control_swap_sv1():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit1 = eng.allocate_qubit()
    qubit2 = eng.allocate_qubit()
    qureg = eng.allocate_qureg(1)
    aws_backend = _awsbraket.AWSBraketBackend(use_hardware=False)
    cmd = Command(eng, Swap, (qubit1, qubit2), controls=qureg)
    assert aws_backend.is_available(cmd)


'''
End of Gate availability Tests
'''


@has_boto3
def test_awsbraket_backend_init():
    backend = _awsbraket.AWSBraketBackend(verbose=True, use_hardware=True)
    assert len(backend._circuit) == 0


@has_boto3
def test_awsbraket_empty_circuit():
    backend = _awsbraket.AWSBraketBackend(verbose=True)
    eng = MainEngine(backend=backend)
    eng.flush()


@has_boto3
def test_awsbraket_invalid_command():
    backend = _awsbraket.AWSBraketBackend(use_hardware=True, verbose=True)
    qb = WeakQubitRef(None, 1)
    cmd = Command(None, gate=SqrtX, qubits=[(qb,)])
    with pytest.raises(Exception):
        backend.receive([cmd])


# ==============================================================================


@has_boto3
def test_awsbraket_sent_error(mocker, sent_error_setup):
    creds, s3_folder, search_value, device_value = sent_error_setup

    var_error = 'ServiceQuotaExceededException'
    mock_boto3_client = mocker.MagicMock(spec=['search_devices', 'get_device', 'create_quantum_task'])
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.create_quantum_task.side_effect = botocore.exceptions.ClientError(
        {"Error": {"Code": var_error, "Message": "Msg error for " + var_error}},
        "create_quantum_task",
    )
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    backend = _awsbraket.AWSBraketBackend(
        verbose=True,
        credentials=creds,
        s3_folder=s3_folder,
        use_hardware=True,
        device='Aspen-8',
        num_runs=10,
    )
    eng = MainEngine(backend=backend, verbose=True)
    qubit = eng.allocate_qubit()
    Rx(0.5) | qubit
    qubit[0].__del__()
    with pytest.raises(botocore.exceptions.ClientError):
        eng.flush()

    # atexit sends another FlushGate, therefore we remove the backend:
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.next_engine = dummy


@has_boto3
def test_awsbraket_sent_error_2():
    backend = _awsbraket.AWSBraketBackend(verbose=True, use_hardware=True, device='Aspen-8')
    eng = MainEngine(
        backend=backend,
        engine_list=[AutoReplacer(DecompositionRuleSet())],
        verbose=True,
    )
    qubit = eng.allocate_qubit()
    Rx(math.pi) | qubit

    with pytest.raises(NoGateDecompositionError):
        SqrtX | qubit
        # no setup to decompose SqrtX gate for Aspen-8,
        # so not accepted by the backend
    dummy = DummyEngine()
    dummy.is_last_engine = True
    eng.next_engine = dummy


# ==============================================================================


@has_boto3
def test_awsbraket_retrieve(mocker, retrieve_setup):
    (arntask, creds, completed_value, device_value, results_dict) = retrieve_setup

    mock_boto3_client = mocker.MagicMock(spec=['get_quantum_task', 'get_device', 'get_object'])
    mock_boto3_client.get_quantum_task.return_value = completed_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.get_object.return_value = results_dict
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    backend = _awsbraket.AWSBraketBackend(retrieve_execution=arntask, credentials=creds, num_retries=2, verbose=True)

    mapper = BasicMapperEngine()
    res = dict()
    for i in range(4):
        res[i] = i
    mapper.current_mapping = res

    eng = MainEngine(backend=backend, engine_list=[mapper], verbose=True)

    separate_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(3)
    del separate_qubit
    eng.flush()

    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[2], qureg[1]])
    assert prob_dict['000'] == 0.04
    assert prob_dict['101'] == 0.2
    assert prob_dict['010'] == 0.8

    # Unknown qubit or no mapper
    invalid_qubit = [WeakQubitRef(eng, 10)]
    with pytest.raises(RuntimeError):
        eng.backend.get_probabilities(invalid_qubit)


# ==============================================================================


@has_boto3
def test_awsbraket_backend_functional_test(mocker, functional_setup, mapper):
    (
        creds,
        s3_folder,
        search_value,
        device_value,
        qtarntask,
        completed_value,
        results_dict,
    ) = functional_setup

    mock_boto3_client = mocker.MagicMock(
        spec=['search_devices', 'get_device', 'create_quantum_task', 'get_quantum_task', 'get_object']
    )
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.create_quantum_task.return_value = qtarntask
    mock_boto3_client.get_quantum_task.return_value = completed_value
    mock_boto3_client.get_object.return_value = results_dict
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    backend = _awsbraket.AWSBraketBackend(
        verbose=True,
        credentials=creds,
        s3_folder=s3_folder,
        use_hardware=True,
        device='Aspen-8',
        num_runs=10,
        num_retries=2,
    )
    # no circuit has been executed -> raises exception
    with pytest.raises(RuntimeError):
        backend.get_probabilities([])

    from projectq.backends import ResourceCounter

    rcount = ResourceCounter()
    engine_list = [rcount]
    if mapper is not None:
        engine_list.append(mapper)
    eng = MainEngine(backend=backend, engine_list=engine_list, verbose=True)

    unused_qubit = eng.allocate_qubit()
    qureg = eng.allocate_qureg(3)

    H | qureg[0]
    S | qureg[1]
    T | qureg[2]
    NOT | qureg[0]
    Y | qureg[1]
    Z | qureg[2]
    Rx(0.1) | qureg[0]
    Ry(0.2) | qureg[1]
    Rz(0.3) | qureg[2]
    R(0.6) | qureg[2]
    C(X) | (qureg[1], qureg[2])
    C(Swap) | (qureg[0], qureg[1], qureg[2])
    H | qureg[0]
    C(Z) | (qureg[1], qureg[2])
    C(R(0.5)) | (qureg[1], qureg[0])
    C(NOT, 2) | ([qureg[2], qureg[1]], qureg[0])
    Swap | (qureg[2], qureg[0])
    Tdag | qureg[1]
    Sdag | qureg[0]

    All(Barrier) | qureg
    del unused_qubit
    # measure; should be all-0 or all-1
    All(Measure) | qureg
    # run the circuit
    eng.flush()

    prob_dict = eng.backend.get_probabilities([qureg[0], qureg[1]])
    assert prob_dict['00'] == pytest.approx(0.84)
    assert prob_dict['01'] == pytest.approx(0.06)

    eng.flush(deallocate_qubits=True)


@has_boto3
def test_awsbraket_functional_test_as_engine(mocker, functional_setup):
    (
        creds,
        s3_folder,
        search_value,
        device_value,
        qtarntask,
        completed_value,
        results_dict,
    ) = functional_setup

    mock_boto3_client = mocker.MagicMock(
        spec=['search_devices', 'get_device', 'create_quantum_task', 'get_quantum_task', 'get_object']
    )
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value
    mock_boto3_client.create_quantum_task.return_value = qtarntask
    mock_boto3_client.get_quantum_task.return_value = completed_value
    mock_boto3_client.get_object.return_value = copy.deepcopy(results_dict)
    mocker.patch('boto3.client', return_value=mock_boto3_client, autospec=True)

    backend = _awsbraket.AWSBraketBackend(
        verbose=True,
        credentials=creds,
        s3_folder=s3_folder,
        use_hardware=True,
        device='Aspen-8',
        num_runs=10,
        num_retries=2,
    )
    # no circuit has been executed -> raises exception
    with pytest.raises(RuntimeError):
        backend.get_probabilities([])

    eng = MainEngine(backend=DummyEngine(save_commands=True), engine_list=[backend], verbose=True)

    unused_qubit = eng.allocate_qubit()  # noqa: F841
    qureg = eng.allocate_qureg(3)

    H | qureg[0]
    CNOT | (qureg[0], qureg[1])
    eng.flush()

    assert len(eng.backend.received_commands) == 7
    assert eng.backend.received_commands[4].gate == H
    assert eng.backend.received_commands[4].qubits[0][0].id == qureg[0].id
    assert eng.backend.received_commands[5].gate == X
    assert eng.backend.received_commands[5].control_qubits[0].id == qureg[0].id
    assert eng.backend.received_commands[5].qubits[0][0].id == qureg[1].id

    # NB: also test that we can call eng.flush() multiple times

    mock_boto3_client.get_object.return_value = copy.deepcopy(results_dict)

    CNOT | (qureg[1], qureg[0])
    H | qureg[1]
    eng.flush()

    assert len(eng.backend.received_commands) == 10
    assert eng.backend.received_commands[7].gate == X
    assert eng.backend.received_commands[7].control_qubits[0].id == qureg[1].id
    assert eng.backend.received_commands[7].qubits[0][0].id == qureg[0].id
    assert eng.backend.received_commands[8].gate == H
    assert eng.backend.received_commands[8].qubits[0][0].id == qureg[1].id

    eng.flush(deallocate_qubits=True)



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setup.awsbraket."""

import pytest
from unittest.mock import patch
import json

# ==============================================================================

_has_boto3 = True
try:
    import projectq.setups.awsbraket

except ImportError:
    _has_boto3 = False

has_boto3 = pytest.mark.skipif(not _has_boto3, reason="boto3 package is not installed")

# ==============================================================================

search_value = {
    "devices": [
        {
            "deviceArn": "arn1",
            "deviceName": "SV1",
            "deviceType": "SIMULATOR",
            "deviceStatus": "ONLINE",
            "providerName": "pname1",
        },
        {
            "deviceArn": "arn2",
            "deviceName": "Aspen-8",
            "deviceType": "QPU",
            "deviceStatus": "OFFLINE",
            "providerName": "pname1",
        },
        {
            "deviceArn": "arn3",
            "deviceName": "IonQ Device",
            "deviceType": "QPU",
            "deviceStatus": "ONLINE",
            "providerName": "pname2",
        },
    ]
}

device_value_devicecapabilities = json.dumps(
    {
        "braketSchemaHeader": {
            "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
            "version": "1",
        },
        "service": {
            "executionWindows": [
                {
                    "executionDay": "Everyday",
                    "windowStartHour": "11:00",
                    "windowEndHour": "12:00",
                }
            ],
            "shotsRange": [1, 10],
            "deviceLocation": "us-east-1",
        },
        "action": {
            "braket.ir.jaqcd.program": {
                "actionType": "braket.ir.jaqcd.program",
                "version": ["1"],
                "supportedOperations": ["H"],
            }
        },
        "paradigm": {
            "qubitCount": 30,
            "nativeGateSet": ["ccnot", "cy"],
            "connectivity": {
                "fullyConnected": False,
                "connectivityGraph": {"1": ["2", "3"]},
            },
        },
        "deviceParameters": {
            "properties": {
                "braketSchemaHeader": {
                    "const": {
                        "name": "braket.device_schema.rigetti.rigetti_device_parameters",
                        "version": "1",
                    }
                }
            },
            "definitions": {
                "GateModelParameters": {
                    "properties": {
                        "braketSchemaHeader": {
                            "const": {
                                "name": "braket.device_schema.gate_model_parameters",
                                "version": "1",
                            }
                        }
                    }
                }
            },
        },
    }
)

device_value = {
    "deviceName": "Aspen-8",
    "deviceType": "QPU",
    "providerName": "provider1",
    "deviceStatus": "OFFLINE",
    "deviceCapabilities": device_value_devicecapabilities,
}

creds = {
    'AWS_ACCESS_KEY_ID': 'aws_access_key_id',
    'AWS_SECRET_KEY': 'aws_secret_key',
}


@has_boto3
@patch('boto3.client')
@pytest.mark.parametrize("var_device", ['SV1', 'Aspen-8', 'IonQ Device'])
def test_awsbraket_get_engine_list(mock_boto3_client, var_device):

    mock_boto3_client.return_value = mock_boto3_client
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value

    engine_list = projectq.setups.awsbraket.get_engine_list(credentials=creds, device=var_device)
    assert len(engine_list) == 12


@has_boto3
@patch('boto3.client')
def test_awsbraket_error(mock_boto3_client):

    mock_boto3_client.return_value = mock_boto3_client
    mock_boto3_client.search_devices.return_value = search_value
    mock_boto3_client.get_device.return_value = device_value

    with pytest.raises(projectq.setups.awsbraket.DeviceOfflineError):
        projectq.setups.awsbraket.get_engine_list(credentials=creds, device='Imaginary')
